/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/poll.h>

#include <compiler.h>
#include <fdctl.h>

/**
 * @brief: Close a file descriptor
 * @param fd: The file descriptor to close
 *
 * Closes a file descriptor handling all possible errors. I bet you didn't
 * know that close(2) could return EINTR.
 *
 * @return 0 on error, 1 on success
*/
int fd_close(int fd)
{
	return _fd_close(fd);
}

/**
 * @brief Configure blocking mode on a file descriptor
 * @param fd: FD to set blocking mode on
 * @param b: Whether to enable or disable blocking mode
 *
 * Configures blocking mode on a file descriptor.
 *
 * @return 0 on error, 1 on success.
 */
int fdctl_block(int fd, int b)
{
	int fl;

	fl = fcntl(fd, F_GETFL);
	if ( fl < 0 )
		return 0;

	if ( b )
		fl &= ~O_NONBLOCK;
	else
		fl |= O_NONBLOCK;

	fl = fcntl(fd, F_SETFL, fl);
	if ( fl < 0 )
		return 0;

	return 1;
}

/**
 * @short Configure close-on-exec mode for a file descriptor
 * @param fd: FD upon which to configure close-on-exec mode
 * @param coe: Whether to enable or disable close-on-exec
 *
 * Allows the fd's close-on-exec flag to be configured.
 *
 * @return 0 on error, 1 on success.
 */
int fdctl_coe(int fd, int coe)
{
	int ret;

	if ( coe )
		coe = FD_CLOEXEC;
	else
		coe = 0;

	ret = fcntl(fd, F_SETFD, coe);
	if ( ret < 0 )
		return 0;

	return 1;
}

/** @internal */
int _fd_read(int fd, void *buf, size_t *sz, int *eof)
{
	size_t len = *sz;
	ssize_t ret;

	*sz = 0;

again:
	ret = read(fd, buf, (len > SSIZE_MAX) ? SSIZE_MAX : len);
	if ( ret < 0 ) {
		if ( errno == EINTR )
			goto again;
		if ( errno == EAGAIN &&
			iowait_single_fd(fd, POLLIN) )
			goto again;
		return 0;
	}

	if ( (size_t)ret == 0 ) {
		*eof = 1;
		return 1;
	}

	*sz += (size_t)ret;

	/* This can happen if a long I/O is interrupted for example
	 * on NFS in soft/interruptable mode in the Linux kernel. Or
	 * on sockets.
	 */
	if ( (size_t)ret < len ) {
		buf += (size_t)ret;
		len -= (size_t)ret;
		goto again;
	}

	return 1;
}

/** @internal */
int _fd_write(int fd, const void *buf, size_t len)
{
	ssize_t ret;

	/* TODO: Handle this case by breaking the write up
	 * in to chunks of SSIZE_MAX.
	 */
	if ( unlikely(len > SSIZE_MAX) )
		return 0;

again:
	ret = write(fd, buf, len);
	if ( ret < 0 ) {
		if ( errno == EINTR )
			goto again;
		if ( errno == EAGAIN &&
			iowait_single_fd(fd, POLLOUT) )
			goto again;
		return 0;
	}

	/* This can happen if a long I/O is interrupted for example
	 * on NFS in soft/interruptable mode in the Linux kernel. Or
	 * on sockets.
	 */
	if ( (size_t)ret < len ) {
		buf += (size_t)ret;
		len -= (size_t)ret;
		goto again;
	}

	return 1;
}

/**
 * @param fd: file descriptor
 * @param buf: data to read
 * @param sz: pointer to size of buffer
 * @param eof: EOF flag, is set to 1 if EOF is hit
 *
 * Call read(2) with given parameters but handle all possible
 * errors. We handle short reads, interrupted calls, fd going
 * O_NONBLOCK under us, and only bail on really unrecoverable
 * errors.
 *
 * @return 0 on unrecoverable error, 1 on success. Even an
 * unrecoverable error may return a non-zero value in sz and
 * hence valid data. If EOF is experienced on the file the
 * eof flag is set to 1.
 */
int fd_read(int fd, void *buf, size_t *sz, int *eof)
{
	return _fd_read(fd, buf, sz, eof);
}

/**
 * @param fd: file descriptor
 * @param buf: data to write
 * @param len: length of data
 *
 * Call write(2) with given parameters but handle all possible
 * errors. We handle short writes, interrupted calls, fd going
 * O_NONBLOCK under us, and only bail on really unrecoverable
 * errors.
 *
 * Failure Modes:
 *  0. return 1: success, all bytes were written to the kernel
 *  1. return 0: len was > SSIZE_MAX
 *  2. return 0: totally unrecoverable failure
 *  3. undefined: any failure mode of iowait_single_fd()
 *  4. undefined: any kernel failure mode
 */
int fd_write(int fd, const void *buf, size_t len)
{
	return _fd_write(fd, buf, len);
}

/** @internal */
int _private _fd_close(int fd)
{
	int ret;

	if ( fd < 0 )
		return 1;
intr:
	ret = close(fd);
	if ( ret && errno == EINTR )
		goto intr;

	return (ret == 0);
}
