/*
 * This file is part of piggy
 * Copyright (c) 2005 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
*/

#include <piggy.h>
#include <nbio.h>
#include <fdctl.h>
#include <lzhl.h>

#define __GNU_SOURCE
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

static SSL_CTX *svr_ctx, *client_ctx;

struct proxy {
	struct nbio io;
	uint8_t *buf, *cur, *end;
	struct proxy *peer;
	SSL *ssl;
	struct lzhl *lzhl;
};

static void hex_dump(uint8_t *tmp, size_t len, size_t llen)
{
	size_t i, j;
	size_t line;

	for(j=0; j<len; j+=line, tmp+=line) {
		if ( j+llen > len ) {
			line=len-j;
		}else{
			line=llen;
		}

		printf("%05x : ", j);

		for(i=0; i<line; i++) {
			if ( isprint(tmp[i]) ) {
				printf("%c", tmp[i]);
			}else{
				printf(".");
			}
		}

		for(; i<llen; i++)
			printf(" ");

		for(i=0; i<line; i++)
			printf(" %02x", tmp[i]);

		printf("\n");
	}
	printf("\n");
}

int ssl_proxy_init(void)
{
	SSL_load_error_strings();
	SSL_library_init();

	svr_ctx = SSL_CTX_new(SSLv3_server_method());
	if ( svr_ctx == NULL )
		return 0;

	client_ctx = SSL_CTX_new(SSLv3_client_method());
	if ( client_ctx == NULL ) {
		SSL_CTX_free(svr_ctx);
		svr_ctx = NULL;
		return 0;
	}

	return 1;
}

int ssl_check_return(struct iothread *t, struct proxy *p, int ret)
{
	int err;

	err = SSL_get_error(p->ssl, ret);
	switch ( err ) {
	case SSL_ERROR_WANT_READ:
		nbio_set_wait(t, &p->io, NBIO_READ);
		break;
	case SSL_ERROR_WANT_WRITE:
		nbio_set_wait(t, &p->io, NBIO_WRITE);
		break;
	default:
		return 0;
	}
	nbio_inactive(t, &p->io);
	return 1;
}

static int alloc_buffer(struct proxy *p)
{
#ifdef PAGE_SIZE
	long pagesz = PAGE_SIZE;
#else
	static long pagesz;

	if ( !pagesz )
		pagesz = sysconf(_SC_PAGESIZE);
#endif

	p->buf = malloc(pagesz);
	if ( p->buf == NULL )
		return 0;

	p->cur = p->buf;
	p->end = p->buf + pagesz;
	return 1;
}

static void downsize_buffer(struct proxy *p, size_t len)
{
	p->end = p->buf + len;
	p->buf = realloc(p->buf, len);
}

static void free_buffer(struct proxy *p)
{
	free(p->buf);
	p->buf = p->cur = p->end = NULL;
}

static void proxy_del(struct iothread *t, struct proxy *p)
{
	/* Tell the other side not to free us */
	if ( p->peer ) {
		p->peer->peer = NULL;
		nbio_del(t, &p->peer->io);
	}

	nbio_del(t, &p->io);
}

static void proxy_dtor(struct iothread *t, struct nbio *io)
{
	struct proxy *p = (struct proxy *)io;
	SSL_free(p->ssl);
	fd_close(p->io.fd);
	free_buffer(p);
	free(p);
}

static void slave_rw(struct iothread *t, struct nbio *io);
static void master_rw(struct iothread *t, struct nbio *io);

static struct nbio_ops slave = {
	.read = slave_rw,
	.write = slave_rw,
	.dtor = proxy_dtor,
};

static struct nbio_ops master = {
	.read = master_rw,
	.write = master_rw,
	.dtor = proxy_dtor,
};

static void slave_rw(struct iothread *t, struct nbio *io)
{
	struct proxy *snd = (struct proxy *)io;
	struct proxy *rcv = snd->peer;
	size_t bufsz;
	int ret;

	bufsz = (rcv->end - rcv->cur);

	ret = SSL_write(snd->ssl, rcv->cur, bufsz);
	if ( ret < 0 ) {
		if ( ssl_check_return(t, rcv, ret) ) {
			free_buffer(rcv);
			return;
		}
		goto die;
	}
	if ( ret == 0 )
		goto die;

	rcv->cur += ret;

	/* Finished sending it now */
	if ( rcv->cur >= rcv->end ) {
		free_buffer(rcv);
		snd->io.ops = &master;
		nbio_set_wait(t, &rcv->io, NBIO_READ|NBIO_WRITE);
		nbio_set_wait(t, &snd->io, NBIO_READ|NBIO_WRITE);
	}

	return;
die:
	if ( ret ) 
		printf("SSL Write: %s\n",
			ERR_error_string(ERR_get_error(), NULL));
	proxy_del(t, snd);
}

static void master_rw(struct iothread *t, struct nbio *io)
{
	struct proxy *rcv = (struct proxy *)io;
	struct proxy *snd = rcv->peer;
	struct sockaddr_in sa;
	socklen_t len;
	int ret = 0;

	if ( !alloc_buffer(rcv) ) {
		/* TODO: Move on to OOM list */
		goto die;
	}

	ret = SSL_read(rcv->ssl, rcv->cur, rcv->end - rcv->cur);
	if ( ret < 0 ) {
		if ( ssl_check_return(t, rcv, ret) ) {
			free_buffer(rcv);
			return;
		}
		goto die;
	}
	if ( ret == 0 )
		goto die;

	getpeername(rcv->io.fd, (struct sockaddr *)&sa, &len);
	printf("Got %u bytes from %s:%u\n", ret,
		inet_ntoa(sa.sin_addr), htons(sa.sin_port));

	if ( (ret - 2) == ntohs(*(uint16_t *)(rcv->buf)) ) {
		size_t inlen, outlen;
		int lret;
		char *out;

		inlen = ret - 2;
		outlen = 4096;
		out = calloc(1, outlen);

		lret = lzhl_decompress(rcv->lzhl, out, &outlen, rcv->buf + 2, &inlen);
		printf(" o Decompressed %u -> %u bytes (ret=%i)\n",
			inlen, outlen, lret);
		hex_dump(out, outlen, 16);
		free(out);
	}else{
		printf(" o Uncompressed frame\n");
		hex_dump(rcv->buf, ret, 16);
	}

	downsize_buffer(rcv, ret);

	snd->io.ops = &slave;
	nbio_set_wait(t, &snd->io, NBIO_WRITE);
	nbio_set_wait(t, &rcv->io, 0);

	return;
die:
	free_buffer(rcv);
	if ( ret )
		printf("SSL Read: %s\n",
				ERR_error_string(ERR_get_error(), NULL));
	proxy_del(t, rcv);
}

static void conn_synchronise(struct iothread *t, struct proxy *p)
{
	p->io.ops = &master;
	if ( p->peer->io.ops == &master ) {
		nbio_set_wait(t, &p->io, NBIO_READ);
		nbio_set_wait(t, &p->peer->io, NBIO_READ);
	}else{
		nbio_set_wait(t, &p->io, 0);
	}
}

static void proxy_accept(struct iothread *t, struct nbio *io)
{
	struct proxy *cl = (struct proxy *)io;
	int ret;

	ret = SSL_accept(cl->ssl);
	switch ( ret ) {
	case 1:
		conn_synchronise(t, cl);
		return;
	case 0:
		break;
	default:
		if ( !ssl_check_return(t, cl, ret) )
			break;
		return;
	}

	printf("SSL accept: %s\n", ERR_error_string(ERR_get_error(), NULL));
	proxy_del(t, cl);
}

static struct nbio_ops accepter = {
	.write = proxy_accept,
	.read = proxy_accept,
	.dtor = proxy_dtor,
};

static void proxy_connect(struct iothread *t, struct nbio *io)
{
	struct proxy *sv = (struct proxy *)io;
	int ret;

	ret = SSL_connect(sv->ssl);
	switch ( ret ) {
	case 1:
		conn_synchronise(t, sv);
		return;
	case 0:
		break;
	default:
		if ( !ssl_check_return(t, sv, ret) )
			break;
		return;
	}

	printf("SSL Connect: %s\n", ERR_error_string(ERR_get_error(), NULL));
	proxy_del(t, sv);
}

static struct nbio_ops connecter = {
	.write = proxy_connect,
	.read = proxy_connect,
	.dtor = proxy_dtor,
};

static void proxy_tcp_conn(struct iothread *t, struct nbio *io)
{
	struct proxy *sv = (struct proxy *)io;
	int err;
	socklen_t len = sizeof(err);

	if ( getsockopt(sv->io.fd, SOL_SOCKET, SO_ERROR, &err, &len) )
		goto die;

	if ( len != sizeof(err) ) {
		goto die;
	}

	if ( err != 0 ) {
		errno = err;
		fprintf(stderr, "deferred connect(): %s\n", strerror(errno));
		goto die;
	}

	sv->io.ops = &connecter;
	proxy_connect(t, &sv->io);
	return;
die:
	proxy_del(t, sv);
}

static struct nbio_ops tcp_connecter = {
	.write = proxy_tcp_conn,
	.dtor = proxy_dtor,
};

void ssl_proxy_new(struct iothread *t, int s, void *priv)
{
	struct proxy_args *arg = priv;
	struct proxy *cl = NULL, *sv = NULL;
	struct sockaddr_in sa;

	cl = calloc(1, sizeof(*cl));
	if ( cl == NULL ) {
		fprintf(stderr, "cl: calloc(): %s\n", strerror(errno));
		goto err;
	}

	cl->ssl = SSL_new(svr_ctx);
	if ( cl->ssl == NULL )
		goto err_free_cl;

	if ( SSL_use_certificate_file(cl->ssl, "test2.crt",
					SSL_FILETYPE_PEM) != 1 ) {
		printf("SSL_CTX_use_certificate_file()\n");
		goto err_free_cl;
	}

	if ( SSL_use_PrivateKey_file(cl->ssl, "test2.key",
					SSL_FILETYPE_PEM) != 1 ) {
		printf("SSL_CTX_use_PrivateKey_file()\n");
		goto err_free_cl;
	}

	if ( SSL_check_private_key(cl->ssl) != 1 ) {
		printf("SSL_check_private_key()\n");
		goto err_free_cl;
	}

	sv = calloc(1, sizeof(*sv));
	if ( sv == NULL ) {
		fprintf(stderr, "sv: calloc(): %s\n", strerror(errno));
		goto err_free_cl_ssl;
	}

	sv->ssl = SSL_new(client_ctx);
	if ( sv->ssl == NULL )
		goto err_free;

	cl->peer = sv;
	cl->io.fd = s;
	if ( !SSL_set_fd(cl->ssl, cl->io.fd) )
		goto err_free_ssl;

	sv->peer = cl;
	sv->io.fd = socket(PF_INET, SOCK_STREAM, 0);
	if ( sv->io.fd < 0 ) {
		fprintf(stderr, "socket(): %s\n", strerror(errno));
		goto err_free_ssl;
	}
	if ( !SSL_set_fd(sv->ssl, sv->io.fd) )
		goto err_close;
	if ( !socket_nonblock(sv->io.fd) ) {
		fprintf(stderr, "socket_nonblock(): %s\n", strerror(errno));
		goto err_close;
	}

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = arg->addr;
	sa.sin_port = arg->port;

	cl->lzhl = lzhl_init();
	sv->lzhl = lzhl_init();

	cl->io.ops = &accepter;
	nbio_add(t, &cl->io, 0);
	proxy_accept(t, &cl->io);

	if ( !connect(sv->io.fd, (struct sockaddr *)&sa, sizeof(sa)) ) {
		nbio_add(t, &sv->io, 0);
		sv->io.ops = &connecter;
		proxy_connect(t, &sv->io);
	}else{
		if ( errno != EINPROGRESS ) {
			fprintf(stderr, "connect(): %s\n", strerror(errno));
			goto err_close;
		}
		sv->io.ops = &tcp_connecter;
		nbio_add(t, &sv->io, NBIO_WRITE);
	}

	return;

err_close:
	fd_close(sv->io.fd);
err_free_ssl:
	SSL_free(sv->ssl);
err_free:
	free(sv);
err_free_cl_ssl:
	SSL_free(cl->ssl);
err_free_cl:
	free(cl);
err:
	fd_close(s);
}
