/*
 * This file is part of piggy
 * Copyright (c) 2005 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
*/

#include <piggy.h>
#include <nbio.h>
#include <fdctl.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

struct proxy {
	struct nbio io;
	uint8_t *buf, *cur, *end;
	struct proxy *peer;
};

static int alloc_buffer(struct proxy *p)
{
#ifdef PAGE_SIZE
	long pagesz = PAGE_SIZE;
#else
	static long pagesz;

	if ( !pagesz )
		pagesz = sysconf(_SC_PAGESIZE);
#endif

	p->buf = malloc(pagesz);
	if ( p->buf == NULL )
		return 0;

	p->cur = p->buf;
	p->end = p->buf + pagesz;
	return 1;
}

static void downsize_buffer(struct proxy *p, size_t len)
{
	p->end = p->buf + len;
	p->buf = realloc(p->buf, len);
}

static void free_buffer(struct proxy *p)
{
	free(p->buf);
	p->buf = p->cur = p->end = NULL;
}

static void proxy_del(struct iothread *t, struct proxy *p)
{
	/* Tell the other side not to free us */
	if ( p->peer ) {
		p->peer->peer = NULL;
		nbio_del(t, &p->peer->io);
	}

	nbio_del(t, &p->io);
}

static void proxy_dtor(struct iothread *t, struct nbio *io)
{
	struct proxy *p = (struct proxy *)io;
	fd_close(p->io.fd);
	free_buffer(p);
	free(p);
}

static void proxy_writable(struct iothread *t, struct nbio *io)
{
	struct proxy *snd = (struct proxy *)io;
	struct proxy *rcv = snd->peer;
	struct sockaddr_in sa;
	socklen_t len;
	size_t bufsz;
	ssize_t ret;

	bufsz = (rcv->end - rcv->cur);

	getpeername(snd->io.fd, (struct sockaddr *)&sa, &len);
	printf("Sending %u bytes to %s:%u\n", bufsz,
		inet_ntoa(sa.sin_addr), htons(sa.sin_port));

	printf("%s\n", rcv->cur);

	ret = send(snd->io.fd, rcv->cur, bufsz, MSG_NOSIGNAL);
	if ( ret < 0 ) {
		if ( errno == EAGAIN ) {
			if ( nbio_get_wait(&snd->io) & NBIO_READ) {
				nbio_set_wait(t, &snd->io, NBIO_READ);
			}else{
				nbio_inactive(t, &snd->io);
			}
			return;
		}
		goto die;
	}
	if ( ret == 0 )
		goto die;

	rcv->cur += ret;
	if ( rcv->cur >= rcv->end ) {
		free_buffer(rcv);
		nbio_set_wait(t, &rcv->io, NBIO_READ);
		nbio_set_wait(t, &snd->io, NBIO_READ);
	}

	return;
die:
	proxy_del(t, snd);
}

static void proxy_readable(struct iothread *t, struct nbio *io)
{
	struct proxy *rcv = (struct proxy *)io;
	struct proxy *snd = rcv->peer;
	struct sockaddr_in sa;
	socklen_t len;
	ssize_t ret;

	if ( !alloc_buffer(rcv) ) {
		/* TODO: Move on to OOM list */
		goto die;
	}

	ret = read(rcv->io.fd, rcv->cur, rcv->end - rcv->cur);
	if ( ret < 0 ) {
		if ( errno == EAGAIN ) {
			if ( nbio_get_wait(&rcv->io) & NBIO_WRITE ) {
				nbio_set_wait(t, &rcv->io, NBIO_WRITE);
			}else{
				nbio_inactive(t, &rcv->io);
			}
			free_buffer(rcv);
			return;
		}
		goto die;
	}
	if ( ret == 0 )
		goto die;

	getpeername(rcv->io.fd, (struct sockaddr *)&sa, &len);
	printf("Got %u bytes from %s:%u\n", ret,
		inet_ntoa(sa.sin_addr), htons(sa.sin_port));

	rcv->buf[ret] = '\0';
	printf("%s\n", rcv->buf);

	downsize_buffer(rcv, ret);

	nbio_set_wait(t, &snd->io, nbio_get_wait(&snd->io) | NBIO_WRITE);
	nbio_set_wait(t, &rcv->io, nbio_get_wait(&rcv->io) & NBIO_WRITE);

	return;
die:
	free_buffer(rcv);
	proxy_del(t, rcv);
}

static struct nbio_ops connected = {
	.read = proxy_readable,
	.write = proxy_writable,
	.dtor = proxy_dtor,
};

static void proxy_conn(struct iothread *t, struct nbio *io)
{
	struct proxy *sv = (struct proxy *)io;
	struct proxy *cl = sv->peer;
	int err;
	socklen_t len = sizeof(err);

	if ( getsockopt(sv->io.fd, SOL_SOCKET, SO_ERROR, &err, &len) )
		goto die;

	if ( len != sizeof(err) ) {
		goto die;
	}

	if ( err != 0 ) {
		errno = err;
		fprintf(stderr, "deferred connect(): %s\n", strerror(errno));
		goto die;
	}

	printf("Connection succeeded\n");

	cl->io.ops = &connected;
	sv->io.ops = &connected;
	nbio_set_wait(t, &sv->io, NBIO_READ);
	nbio_add(t, &cl->io, NBIO_READ);
	return;
die:
	proxy_del(t, sv);
}

static struct nbio_ops connector = {
	.write = proxy_conn,
	.dtor = proxy_dtor,
};

void proxy_new(struct iothread *t, int s, void *priv)
{
	struct proxy_args *arg = priv;
	struct proxy *cl = NULL, *sv = NULL;
	struct sockaddr_in sa;

	cl = calloc(1, sizeof(*cl));
	if ( cl == NULL ) {
		fprintf(stderr, "cl: calloc(): %s\n", strerror(errno));
		goto err;
	}

	sv = calloc(1, sizeof(*sv));
	if ( sv == NULL ) {
		fprintf(stderr, "sv: calloc(): %s\n", strerror(errno));
		goto err_free_cl;
	}

	cl->io.fd = s;
	cl->peer = sv;

	sv->peer = cl;
	sv->io.fd = socket(PF_INET, SOCK_STREAM, 0);
	if ( sv->io.fd < 0 ) {
		fprintf(stderr, "socket(): %s\n", strerror(errno));
		goto err_free;
	}
	if ( !socket_nonblock(sv->io.fd) ) {
		fprintf(stderr, "socket_nonblock(): %s\n", strerror(errno));
		goto err_free;
	}

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = arg->addr;
	sa.sin_port = arg->port;

	if ( !connect(sv->io.fd, (struct sockaddr *)&sa, sizeof(sa)) ) {
		cl->io.ops = &connected;
		sv->io.ops = &connected;
		nbio_add(t, &sv->io, NBIO_READ);
		nbio_add(t, &cl->io, NBIO_READ);
	}else{
		if ( errno != EINPROGRESS ) {
			fprintf(stderr, "connect(): %s\n", strerror(errno));
			goto err_close;
		}
		sv->io.ops = &connector;
		nbio_add(t, &sv->io, NBIO_WRITE);
	}

	return;

err_close:
	fd_close(sv->io.fd);
err_free:
	free(sv);
err_free_cl:
	free(cl);
err:
	fd_close(s);
}
