/*
 * This file is part of piggy
 * Copyright (c) 2005 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h> /* errno */
#include <sys/socket.h>  /* inet_aton(3) */
#include <netinet/in.h> /* inet_aton(3)/htons(3) */
#include <arpa/inet.h> /* inet_aton(3) */

#include <compiler.h>
#include <fdctl.h>
#include <bytesex.h>
#include <list.h>
#include <nbio.h>

static LIST_HEAD(sweepers);
static const char *cmd;
static struct iothread iothread;
static int random_fd;
#define RANDOM_BUF_LEN 1024
static uint32_t random_buf[RANDOM_BUF_LEN];
static unsigned int ridx;

struct sweep {
	struct nbio io;
	struct sockaddr_in sa;
	int outfd;
	struct list_head sweep_list;
};

static int init_random(void)
{
	random_fd = open("/dev/urandom", O_RDONLY);

	if ( random_fd < 0 ) {
		perror("open() on /dev/urandom");
		return 0;
	}

	return 1;
}

static uint32_t get_random_address(void)
{
	if ( (ridx % RANDOM_BUF_LEN) == 0 ) {
		void *buf = random_buf;
		size_t len = sizeof(random_buf);
		ssize_t ret;

again:
		ret = read(random_fd, buf, len);
		if ( ret < 0 )
			goto err;
		if ( ret == 0 )
			goto err;

		if ( len > (size_t)ret ) {
			buf += ret;
			len -= ret;
			goto again;
		}
	}

	return random_buf[ridx++ % RANDOM_BUF_LEN];
err:
	abort();
}

static void sweeper_dtor(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;
	fd_close(s->io.fd);
	fd_close(s->outfd);
	
	list_add(&s->sweep_list, &sweepers);
}

void finger_read(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;
	char buf[1024];
	char *ptr = buf;
	size_t len = sizeof(buf);
	ssize_t ret;

again:
	ret = read(s->io.fd, ptr, len);
	if ( ret < 0 ) {
		if ( errno == EAGAIN ) {
			nbio_inactive(t, io);
		}else{
			perror("read()");
			nbio_del(t, io);
		}
	}else if ( ret == 0 ) {
		nbio_del(t, io);
	}else if ( (size_t)ret < len ) {
		ptr += ret;
		len -= ret;
		goto again;
	}

	len = sizeof(buf) - len;
	buf[len] = '\0';

	if ( len == 0 )
		return;

	printf("Got %s\n", buf);

	if ( s->outfd < 0 ) {
		char fn[128];
		snprintf(fn, sizeof(fn), "%s_%u.log",
				inet_ntoa(s->sa.sin_addr),
				htons(s->sa.sin_port));
		s->outfd = open(fn, O_WRONLY|O_CREAT|O_EXCL, 0600);
		if ( s->outfd < 0 ) {
			perror("open()");
			nbio_del(t, io);
			return;
		}
	}

	if ( !fd_write(s->outfd, buf, len) ) {
		perror("write()");
		fd_close(s->outfd);
		s->outfd = -1;
	}
}

void finger_write(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;
	ssize_t ret;

again:
	ret = write(s->io.fd, "\n", 1);
	if ( ret < 0 ) {
		if ( errno == EAGAIN ) {
			nbio_inactive(t, io);
		}else{
			perror("write()");
			nbio_del(t, io);
		}
		return;
	}

	if ( ret == 0 )
		goto again;

	printf("Wrote a byte\n");
	nbio_set_wait(t, io, NBIO_READ);
}

static struct nbio_ops finger = {
	.read = finger_read,
	.write = finger_write,
	.dtor = sweeper_dtor,
};

void banner_read(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;
	char buf[1024];
	char *ptr = buf;
	size_t len = sizeof(buf);
	ssize_t ret;

again:
	ret = read(s->io.fd, ptr, len);
	if ( ret < 0 ) {
		if ( errno == EAGAIN ) {
			nbio_inactive(t, io);
		}else{
			perror("read()");
			nbio_del(t, io);
		}
	}else if ( ret == 0 ) {
		nbio_del(t, io);
	}else if ( (size_t)ret < len ) {
		ptr += ret;
		len -= ret;
		goto again;
	}

	len = sizeof(buf) - len;
	buf[len] = '\0';

	if ( len == 0 )
		return;

	printf("banner: %s: %s\n", inet_ntoa(s->sa.sin_addr), buf);
	nbio_del(t, io);
}

static struct nbio_ops banner = {
	.read = banner_read,
	.dtor = sweeper_dtor,
};

static void sweeper_connected(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;

	printf("Connection succeeded: %s %u\n",
		inet_ntoa(s->sa.sin_addr),
		ntohs(s->sa.sin_port));

	//s->io.ops = &finger;
	//nbio_set_wait(t, io, NBIO_READ|NBIO_WRITE);
	s->io.ops = &banner;
	nbio_set_wait(t, io, NBIO_READ);
}

static void sweeper_connwrite(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;
	int err;
	socklen_t len = sizeof(err);

	if ( getsockopt(s->io.fd, SOL_SOCKET, SO_ERROR, &err, &len) )
		goto die;

	if ( len != sizeof(err) ) {
		goto die;
	}

	if ( err != 0 ) {
		fprintf(stderr, "deferred connect(): %s\n", strerror(err));
		goto die;
	}

	sweeper_connected(t, &s->io);
	return;
die:
	nbio_del(t, &s->io);
}

static struct nbio_ops connector = {
	.write = sweeper_connwrite,
	.dtor = sweeper_dtor,
};

static void sweeper_connect(struct iothread *t, struct nbio *io)
{
	struct sweep *s = (struct sweep *)io;

again:
	s->io.fd = socket(PF_INET, SOCK_STREAM, 0);
	if ( s->io.fd < 0 ) {
		perror("socket()");
		return;
	}

	if ( !socket_nonblock(s->io.fd) ) {
		fd_close(s->io.fd);
		return;
	}

	s->sa.sin_addr.s_addr = get_random_address();
//	printf("Trying %s\n", inet_ntoa(s->sa.sin_addr));

	if ( !connect(s->io.fd, (struct sockaddr *)&s->sa, sizeof(s->sa)) ) {
		/* Add any old io_ops as long as it has dtor */
		s->io.ops = &connector;
		list_del(&s->sweep_list);
		nbio_add(t, &s->io, 0);
		sweeper_connected(t, &s->io);
	}else{
		if ( errno != EINPROGRESS ) {
			perror("connect()");
			fd_close(s->io.fd);
			goto again;
		}
		s->io.ops = &connector;
		list_del(&s->sweep_list);
		nbio_add(t, &s->io, NBIO_WRITE);
		nbio_inactive(t, &s->io);
	}

	return;
}

static int sweeper_new(struct iothread *t, int port)
{
	struct sweep *s;

	s = calloc(1, sizeof(*s));
	if ( s == NULL ) {
		perror("calloc()");
		return 0;
	}

	s->sa.sin_family = AF_INET;
	s->sa.sin_port = htons(port);
	s->outfd = -1;

	INIT_LIST_HEAD(&s->sweep_list);
	list_add_tail(&s->sweep_list, &sweepers);

	return 1;
}

int main(int argc, char **argv)
{
	const char *plugin = NULL;
	struct sweep *n, *tmp;
	int i;
	setlinebuf(stdout);

	cmd = argv[0];

	sig_pipeignore();

	if ( !init_random() )
		return EXIT_FAILURE;

	if ( !nbio_init(&iothread, plugin) )
		return EXIT_FAILURE;

	printf("%s: Using %s eventloop plugin\n",
		cmd, iothread.plugin->name);
	printf("%s: Working on port %i, concurrency %i\n",
		cmd, atoi(argv[1]), atoi(argv[2]));

	for(i=0; i < atoi(argv[2]); i++)
		sweeper_new(&iothread, atoi(argv[1]));

	do {
		list_for_each_entry_safe(n, tmp, &sweepers, sweep_list)
			sweeper_connect(&iothread, &n->io);
		nbio_pump(&iothread);
	}while ( !list_empty(&iothread.active) );
	

	nbio_fini(&iothread);

	return EXIT_SUCCESS;
}
